from naas_abi_core.models.OpenRouter import ChatOpenRouter


def test_openai():
    llm = ChatOpenRouter(model_name="openai/gpt-4.1")
    result = llm.invoke("What is the capital of France?").content
    assert result is not None, result
    assert "Paris" in result, result


def test_anthropic():
    llm = ChatOpenRouter(model_name="anthropic/claude-sonnet-4.5")
    result = llm.invoke("What is the capital of France?").content
    assert result is not None, result
    assert "Paris" in result, result


def test_xai():
    llm = ChatOpenRouter(model_name="x-ai/grok-4")
    result = llm.invoke("What is the capital of France?").content
    assert result is not None, result
    assert "Paris" in result, result


def test_mistral():
    llm = ChatOpenRouter(model_name="mistralai/mistral-large")
    result = llm.invoke("What is the capital of France?").content
    assert result is not None, result
    assert "Paris" in result, result


def test_gemini():
    llm = ChatOpenRouter(model_name="google/gemini-2.5-flash")
    result = llm.invoke("What is the capital of France?").content
    assert result is not None, result
    assert "Paris" in result, result



import uuid
from enum import Enum
from queue import Empty, Queue
from threading import Event, Lock, Thread
from typing import Callable, List, Optional

from naas_abi_core import logger


class JobStatus(Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    RESULT_FETCHED = "result_fetched"


class Job:
    def __init__(self, queue: Optional[Queue], func: Callable, *args, **kwargs):
        self.id = str(uuid.uuid4())
        self.queue = queue
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.status = JobStatus.PENDING
        self.result = None
        self.error = None
        self._completion_event = Event()
        self._lock = Lock()  # For thread-safe status updates

    def execute(self):
        """Execute the job - called by worker thread"""
        with self._lock:
            self.status = JobStatus.RUNNING

        try:
            self.result = self.func(*self.args, **self.kwargs)
            with self._lock:
                self.status = JobStatus.COMPLETED
        except Exception as e:
            self.error = e
            import traceback

            logger.error(f"Job {self.id} failed: {e}\n{traceback.format_exc()}")
            with self._lock:
                self.status = JobStatus.FAILED
        finally:
            self._completion_event.set()
            if self.queue:
                self.queue.put(self)

    def wait(self, timeout: Optional[float] = None) -> bool:
        """Wait for the job to complete"""
        return self._completion_event.wait(timeout)

    def is_completed(self) -> bool:
        """Check if the job is completed (successfully or with failure)"""
        return self.status in (JobStatus.COMPLETED, JobStatus.FAILED)

    def get_result(self):
        """Get the job result, raising any error that occurred"""
        with self._lock:
            self.status = JobStatus.RESULT_FETCHED

        if self.error:
            raise self.error

        return self.result


class WorkerPool:
    def __init__(self, num_workers: int):
        self.job_queue: Queue[Job] = Queue()
        self.workers: List[Thread] = []
        self.shutdown_event = Event()

        # Start worker threads
        for _ in range(num_workers):
            worker = Thread(target=self._worker_loop)
            worker.daemon = True
            worker.start()
            self.workers.append(worker)

    def _worker_loop(self):
        while not self.shutdown_event.is_set():
            try:
                # Get job from queue with timeout to check shutdown periodically
                job = self.job_queue.get(timeout=1.0)
                job.execute()
                self.job_queue.task_done()
            except Empty:
                continue

    def submit(self, job: Job):
        """Submit a job to the worker pool"""
        self.job_queue.put(job)

    def submit_all(self, jobs: List[Job]) -> Queue[Job]:
        """Submit multiple jobs to the worker pool"""
        queue: Queue[Job] = Queue(maxsize=len(jobs))

        for job in jobs:
            if job.queue is None:
                job.queue = queue
            self.submit(job)

        return queue

    def shutdown(self):
        """Shutdown the worker pool"""
        logger.debug("Shutting down worker pool")
        self.shutdown_event.set()
        for worker in self.workers:
            worker.join()



import os
import sys
from io import StringIO
from typing import List

import yaml
from jinja2 import Template
from naas_abi_core import logger
from naas_abi_core.engine.engine_configuration.EngineConfiguration_Deploy import (
    DeployConfiguration,
)
from naas_abi_core.engine.engine_configuration.EngineConfiguration_ObjectStorageService import (
    ObjectStorageServiceConfiguration,
)
from naas_abi_core.engine.engine_configuration.EngineConfiguration_SecretService import (
    SecretServiceConfiguration,
)
from naas_abi_core.engine.engine_configuration.EngineConfiguration_TripleStoreService import (
    TripleStoreServiceConfiguration,
)
from naas_abi_core.engine.engine_configuration.EngineConfiguration_VectorStoreService import (
    VectorStoreServiceConfiguration,
)
from naas_abi_core.services.secret.Secret import Secret
from pydantic import BaseModel, model_validator
from rich.prompt import Prompt
from typing_extensions import Literal, Self


class ServicesConfiguration(BaseModel):
    object_storage: ObjectStorageServiceConfiguration
    triple_store: TripleStoreServiceConfiguration
    vector_store: VectorStoreServiceConfiguration
    secret: SecretServiceConfiguration


class ApiConfiguration(BaseModel):
    title: str = "ABI API"
    description: str = "API for ABI, your Artifical Business Intelligence"
    logo_path: str = "assets/logo.png"
    favicon_path: str = "assets/favicon.ico"
    cors_origins: List[str] = ["http://localhost:9879"]


class FirstPassConfiguration(BaseModel):
    """This is a first pass configuration that is used to load the secret service.

    It is used to load the secret service before the other services are loaded.
    This is because the secret service needs to be loaded before the other services
    are loaded to be able to resolve the secrets.
    This is a first pass configuration that is used to load the secret service.
    """

    class FirstPassServicesConfiguration(BaseModel):
        secret: SecretServiceConfiguration

    services: FirstPassServicesConfiguration


class ModuleConfig(BaseModel):
    path: str | None = None
    module: str | None = None
    enabled: bool
    config: dict = {}

    @model_validator(mode="after")
    def validate_path_or_module(self):
        if self.path is None and self.module is None:
            raise ValueError("Either path or module must be provided")

        if self.path is None:
            assert self.module is not None, (
                "module must be provided if path is not provided"
            )
        if self.module is None:
            assert self.path is not None, (
                "path must be provided if module is not provided"
            )
        return self


class GlobalConfig(BaseModel):
    ai_mode: Literal["cloud", "local", "airgap"]


class EngineConfiguration(BaseModel):
    api: ApiConfiguration

    deploy: DeployConfiguration | None = None

    services: ServicesConfiguration

    global_config: GlobalConfig

    modules: List[ModuleConfig]

    def ensure_default_modules(self) -> None:
        if not any(
            m.path == "naas_abi_core.modules.templatablesparqlquery"
            or m.module == "naas_abi_core.modules.templatablesparqlquery"
            for m in self.modules
        ):
            self.modules.append(
                ModuleConfig(
                    module="naas_abi_core.modules.templatablesparqlquery",
                    enabled=True,
                    config={},
                )
            )

    @model_validator(mode="after")
    def validate_modules(self) -> Self:
        self.ensure_default_modules()
        return self

    @classmethod
    def from_yaml(cls, yaml_path: str) -> "EngineConfiguration":
        with open(yaml_path, "r") as file:
            return cls.from_yaml_content(file.read())

    @classmethod
    def from_yaml_content(cls, yaml_content: str) -> "EngineConfiguration":
        # First we do a pass with the minimal configuration to load the secret service.
        class SecretServiceWrapper:
            secret_service: Secret | None = None

            def __init__(self, secret_service: Secret | None = None):
                self.secret_service = secret_service

            def __getattr__(self, name):
                if self.secret_service is None:
                    return 0
                secret = self.secret_service.get(name)
                if secret is None:
                    if not sys.stdin.isatty():
                        raise ValueError(
                            f"Secret '{name}' not found and no TTY available to prompt. Please provide it via the configured secret service or environment."
                        )
                    value = Prompt.ask(
                        f"[bold yellow]Secret '{name}' not found.[/bold yellow] Please enter the value for [cyan]{name}[/cyan]",
                        password=False,
                    )
                    self.secret_service.set(name, value)
                    return value
                return secret

            def get(self, key, default=None):
                return 0

        first_pass_data = yaml.safe_load(
            StringIO(Template(yaml_content).render(secret=SecretServiceWrapper()))
        )

        first_pass_configuration = FirstPassConfiguration(**first_pass_data)
        secret_service = first_pass_configuration.services.secret.load()

        # Here we can now template the yaml by using `yaml_content` and the secret service.
        # Using Jinja2 template engine.

        logger.debug("Yaml content: {yaml_content}")

        template = Template(yaml_content)
        templated_yaml = template.render(secret=SecretServiceWrapper(secret_service))

        logger.debug(f"Templated yaml: {templated_yaml}")

        data = yaml.safe_load(StringIO(templated_yaml))

        logger.debug(f"Data: {data}")

        return cls(**data)

    @classmethod
    def load_configuration(
        cls, configuration_yaml: str | None = None
    ) -> "EngineConfiguration":
        # This is useful for testing.
        if configuration_yaml is not None:
            return cls.from_yaml_content(configuration_yaml)

        logger.debug(f"Loading configuration from {os.getenv('ENV')}")

        if os.path.exists(f"config.{os.getenv('ENV')}.yaml"):
            return cls.from_yaml(f"config.{os.getenv('ENV')}.yaml")
        elif os.path.exists("config.yaml"):
            return cls.from_yaml("config.yaml")
        else:
            raise FileNotFoundError(
                "Configuration file not found. Please create a config.yaml file or config.{env}.yaml file."
            )


if __name__ == "__main__":
    config = EngineConfiguration.load_configuration()
    print(config)



from urllib.parse import quote
import hashlib
import uuid
from uuid import UUID
import unicodedata
import re

def create_id_from_string(string: str) -> str:
    """Create a URL-safe ID from a string.

    Args:
        string (str): Input string to convert

    Returns:
        str: URL-safe string with spaces replaced by hyphens
    """
    return quote(string.replace(" ", "-").lower(), safe="")

def string_to_uuid(input_string: str) -> UUID:
    """Convert a string to a UUID.

    Args:
        input_string (str): Input string to convert

    Returns:
        str: UUID
    """
    hash_object = hashlib.sha256(input_string.encode())
    hex_hash = hash_object.hexdigest()
    uuid_obj = uuid.UUID(hex_hash[:32])
    return uuid_obj

def create_hash_from_string(input_string: str) -> str:
    """Convert a string to a UUID.

    Args:
        input_string (str): Input string to convert

    Returns:
        str: UUID
    """
    hash_object = hashlib.sha256(input_string.encode())
    hex_hash = hash_object.hexdigest()
    return hex_hash

def normalize(text: str) -> str:
    """Normalize text by removing accents, punctuation, and converting to lowercase."""
    text = text.lower()
    text = unicodedata.normalize('NFKD', text)
    text = re.sub(r'[^\w\s]', '', text)
    text = ''.join(c for c in text if not unicodedata.combining(c))
    return text



import uuid

import pytest
import requests
from naas_abi_core.services.triple_store.adaptors.secondary.Oxigraph import Oxigraph
from rdflib import RDF, BNode, Graph, Literal, URIRef

# class TestOxigraph:
#     """Test suite for the Oxigraph triple store adapter."""

#     @pytest.fixture
#     def mock_oxigraph_url(self):
#         """Mock Oxigraph URL for testing."""
#         return "http://localhost:7878"

#     @pytest.fixture
#     def sample_graph(self):
#         """Create a sample RDF graph for testing."""
#         g = Graph()
#         g.bind("ex", "http://example.org/")

#         subject = URIRef("http://example.org/alice")
#         g.add((subject, RDF.type, URIRef("http://example.org/Person")))
#         g.add((subject, URIRef("http://example.org/name"), Literal("Alice")))
#         g.add((subject, URIRef("http://example.org/age"), Literal(30)))

#         return g

#     @pytest.fixture
#     def mock_successful_response(self):
#         """Mock successful HTTP response for SELECT queries."""
#         response = Mock()
#         response.status_code = 200
#         response.headers = {"Content-Type": "application/sparql-results+json"}
#         response_text = """{
#             "head": {"vars": ["s", "p", "o"]},
#             "results": {
#                 "bindings": [{
#                     "s": {"type": "uri", "value": "http://example.org/alice"},
#                     "p": {"type": "uri", "value": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"},
#                     "o": {"type": "uri", "value": "http://example.org/Person"}
#                 }]
#             }
#         }"""
#         response.text = response_text
#         response.content = response_text.encode('utf-8')
#         response.raise_for_status = Mock()
#         return response

#     @pytest.fixture
#     def mock_turtle_response(self):
#         """Mock Turtle response for CONSTRUCT queries."""
#         response = Mock()
#         response.status_code = 200
#         response.headers = {"Content-Type": "text/turtle"}
#         response.text = """@prefix ex: <http://example.org/> .
# ex:alice a ex:Person ."""
#         response.raise_for_status = Mock()
#         return response

#     def test_init_default_parameters(self):
#         """Test Oxigraph initialization with default parameters."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             adapter = Oxigraph()

#             assert adapter.oxigraph_url == "http://localhost:7878"
#             assert adapter.query_endpoint == "http://localhost:7878/query"
#             assert adapter.update_endpoint == "http://localhost:7878/update"
#             assert adapter.store_endpoint == "http://localhost:7878/store"
#             assert adapter.timeout == 60

#     def test_init_custom_parameters(self, mock_oxigraph_url):
#         """Test Oxigraph initialization with custom parameters."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             custom_url = "http://oxigraph.example.com:8080"
#             adapter = Oxigraph(
#                 oxigraph_url=custom_url,
#                 timeout=30
#             )

#             assert adapter.oxigraph_url == custom_url
#             assert adapter.query_endpoint == f"{custom_url}/query"
#             assert adapter.timeout == 30

#     def test_connection_test_success(self, mock_oxigraph_url):
#         """Test successful connection test during initialization."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             Oxigraph(mock_oxigraph_url)

#             # Verify connection test was performed
#             mock_get.assert_called_once()
#             call_args = mock_get.call_args
#             assert call_args[0][0] == f"{mock_oxigraph_url}/query"
#             assert "query" in call_args[1]["params"]

#     def test_connection_test_failure(self, mock_oxigraph_url):
#         """Test connection failure during initialization."""
#         with patch('requests.get') as mock_get:
#             mock_get.side_effect = requests.exceptions.ConnectionError("Connection failed")

#             with pytest.raises(requests.exceptions.ConnectionError):
#                 Oxigraph(mock_oxigraph_url)

#     def test_insert_success(self, mock_oxigraph_url, sample_graph):
#         """Test successful triple insertion."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)
#             adapter.insert(sample_graph)

#             # Verify the POST call was made correctly
#             assert mock_post.call_count == 1
#             call_args = mock_post.call_args
#             assert call_args[0][0] == adapter.update_endpoint
#             assert call_args[1]["headers"]["Content-Type"] == "application/sparql-update"
#             assert "INSERT DATA" in call_args[1]["data"]
#             assert "alice" in call_args[1]["data"]  # Check that our sample data is there

#     def test_remove_success(self, mock_oxigraph_url, sample_graph):
#         """Test successful triple removal."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)
#             adapter.remove(sample_graph)

#             # Verify the POST call was made correctly
#             assert mock_post.call_count == 1
#             call_args = mock_post.call_args
#             assert call_args[0][0] == adapter.update_endpoint
#             assert call_args[1]["headers"]["Content-Type"] == "application/sparql-update"
#             assert "DELETE DATA" in call_args[1]["data"]
#             assert "alice" in call_args[1]["data"]

#     def test_get_success(self, mock_oxigraph_url, mock_turtle_response):
#         """Test successful retrieval of all triples."""
#         with patch('requests.get') as mock_get:

#             # First call for connection test, second for get
#             mock_get.side_effect = [
#                 Mock(status_code=200, raise_for_status=Mock()),  # Connection test
#                 mock_turtle_response  # Get response
#             ]

#             adapter = Oxigraph(mock_oxigraph_url)
#             result = adapter.get()

#             assert isinstance(result, Graph)
#             # The mock response contains one triple
#             assert len(result) > 0

#     def test_query_select_success(self, mock_oxigraph_url, mock_successful_response):
#         """Test successful SELECT query execution."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value = mock_successful_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             query = "SELECT ?s ?p ?o WHERE { ?s ?p ?o }"
#             result = adapter.query(query)

#             # Verify the query was executed correctly
#             call_args = mock_post.call_args
#             assert call_args[0][0] == adapter.query_endpoint
#             assert call_args[1]["data"] == query
#             assert "sparql-results" in call_args[1]["headers"]["Accept"]

#             # Verify we can iterate results
#             result_list = list(result)
#             assert len(result_list) == 1

#     def test_query_construct_success(self, mock_oxigraph_url, mock_turtle_response):
#         """Test successful CONSTRUCT query execution."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_turtle_response.headers = {"Content-Type": "application/n-triples"}
#             mock_turtle_response.text = "<http://example.org/alice> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://example.org/Person> .\n"
#             mock_post.return_value = mock_turtle_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             query = "CONSTRUCT { ?s ?p ?o } WHERE { ?s ?p ?o }"
#             result = adapter.query(query)

#             # For CONSTRUCT queries, result should be a Graph
#             assert isinstance(result, Graph)

#     def test_query_update_success(self, mock_oxigraph_url):
#         """Test successful UPDATE query execution."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             update_response = Mock()
#             update_response.status_code = 200
#             update_response.headers = {"Content-Type": "text/plain"}
#             update_response.raise_for_status = Mock()
#             mock_post.return_value = update_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             query = "INSERT DATA { <http://example.org/test> <http://example.org/prop> 'value' }"
#             adapter.query(query)

#             # Verify the update was executed correctly
#             call_args = mock_post.call_args
#             assert call_args[0][0] == adapter.update_endpoint
#             assert call_args[1]["data"] == query

#     def test_get_subject_graph_success(self, mock_oxigraph_url, mock_turtle_response):
#         """Test successful subject graph retrieval."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_turtle_response.headers = {"Content-Type": "application/n-triples"}
#             mock_turtle_response.text = "<http://example.org/alice> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://example.org/Person> .\n"
#             mock_post.return_value = mock_turtle_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             subject = URIRef("http://example.org/alice")
#             result = adapter.get_subject_graph(subject)

#             assert isinstance(result, Graph)
#             # Verify the query was constructed correctly
#             call_args = mock_post.call_args
#             assert str(subject) in call_args[1]["data"]
#             assert "CONSTRUCT" in call_args[1]["data"]

#     def test_query_view_delegates_to_query(self, mock_oxigraph_url):
#         """Test that query_view delegates to query method."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)

#             # Mock the query method
#             adapter.query = Mock(return_value="mocked_result")

#             query = "SELECT ?s WHERE { ?s ?p ?o }"
#             result = adapter.query_view("some_view", query)

#             # Verify query was called with the same query string
#             adapter.query.assert_called_once_with(query)
#             assert result == "mocked_result"

#     def test_handle_view_event_no_op(self, mock_oxigraph_url):
#         """Test that handle_view_event is a no-op."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)

#             # This should not raise any exception
#             adapter.handle_view_event(
#                 view=(None, None, None),
#                 event=OntologyEvent.INSERT,
#                 triple=(URIRef("http://example.org/s"), URIRef("http://example.org/p"), URIRef("http://example.org/o"))
#             )

#     def test_insert_http_error(self, mock_oxigraph_url, sample_graph):
#         """Test that HTTP errors are properly raised during insert."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value.raise_for_status.side_effect = requests.exceptions.HTTPError("Server Error")

#             adapter = Oxigraph(mock_oxigraph_url)

#             with pytest.raises(requests.exceptions.HTTPError):
#                 adapter.insert(sample_graph)

#     def test_query_timeout_error(self, mock_oxigraph_url):
#         """Test that timeout errors are properly handled."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.side_effect = requests.exceptions.Timeout("Request timed out")

#             adapter = Oxigraph(mock_oxigraph_url)

#             with pytest.raises(requests.exceptions.Timeout):
#                 adapter.query("SELECT ?s WHERE { ?s ?p ?o }")

#     def test_query_unexpected_content_type(self, mock_oxigraph_url):
#         """Test handling of unexpected content types in query responses."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             unexpected_response = Mock()
#             unexpected_response.status_code = 200
#             unexpected_response.headers = {"Content-Type": "text/html"}
#             unexpected_response.raise_for_status = Mock()
#             mock_post.return_value = unexpected_response

#             adapter = Oxigraph(mock_oxigraph_url)

#             with pytest.raises(ValueError, match="Unexpected content type"):
#                 adapter.query("SELECT ?s WHERE { ?s ?p ?o }")

#     def test_insert_empty_graph(self, mock_oxigraph_url):
#         """Test that inserting an empty graph is handled gracefully."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)
#             empty_graph = Graph()

#             # Insert empty graph - should not make HTTP request
#             adapter.insert(empty_graph)

#             # Verify no POST requests were made for empty graph
#             mock_post.assert_not_called()

#     def test_remove_empty_graph(self, mock_oxigraph_url):
#         """Test that removing from an empty graph works correctly."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value.raise_for_status = Mock()

#             adapter = Oxigraph(mock_oxigraph_url)
#             empty_graph = Graph()

#             # Remove empty graph - should still work
#             adapter.remove(empty_graph)

#             # Verify POST was called (empty DELETE DATA query is still valid)
#             assert mock_post.call_count == 1
#             call_args = mock_post.call_args
#             assert "DELETE DATA" in call_args[1]["data"]

#     def test_query_ask_query(self, mock_oxigraph_url):
#         """Test ASK query handling."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             ask_response = Mock()
#             ask_response.status_code = 200
#             ask_response.headers = {"Content-Type": "application/sparql-results+json"}
#             ask_response.text = """{
#                 "head": {},
#                 "boolean": true
#             }"""
#             ask_response.raise_for_status = Mock()
#             mock_post.return_value = ask_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             query = "ASK WHERE { ?s ?p ?o }"
#             result = adapter.query(query)

#             # ASK queries return results but with different structure
#             result_list = list(result)
#             # For ASK queries, the result structure is different but should still be iterable
#             assert isinstance(result_list, list)

#     def test_query_with_different_datatypes(self, mock_oxigraph_url):
#         """Test query results with different RDF datatypes."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             # Mock response with various datatypes
#             response = Mock()
#             response.status_code = 200
#             response.headers = {"Content-Type": "application/sparql-results+json"}
#             response.text = """{
#                 "head": {"vars": ["uri", "integer", "string", "date"]},
#                 "results": {
#                     "bindings": [{
#                         "uri": {"type": "uri", "value": "http://example.org/test"},
#                         "integer": {"type": "literal", "value": "42", "datatype": "http://www.w3.org/2001/XMLSchema#integer"},
#                         "string": {"type": "literal", "value": "test string"},
#                         "date": {"type": "literal", "value": "2023-01-01", "datatype": "http://www.w3.org/2001/XMLSchema#date"}
#                     }]
#                 }
#             }"""
#             response.raise_for_status = Mock()
#             mock_post.return_value = response

#             adapter = Oxigraph(mock_oxigraph_url)
#             result = adapter.query("SELECT ?uri ?integer ?string ?date WHERE { ?s ?p ?o }")

#             rows = list(result)
#             assert len(rows) == 1
#             row = rows[0]

#             # Test that different datatypes are handled correctly
#             assert str(row.uri) == "http://example.org/test"
#             assert isinstance(row.integer, Literal)
#             assert int(row.integer) == 42
#             assert str(row.string) == "test string"

#     def test_get_subject_graph_empty_result(self, mock_oxigraph_url):
#         """Test get_subject_graph when no triples are found for the subject."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             # Mock empty CONSTRUCT result
#             empty_response = Mock()
#             empty_response.status_code = 200
#             empty_response.headers = {"Content-Type": "application/n-triples"}
#             empty_response.text = ""  # Empty N-Triples
#             empty_response.raise_for_status = Mock()
#             mock_post.return_value = empty_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             subject = URIRef("http://example.org/nonexistent")
#             result = adapter.get_subject_graph(subject)

#             assert isinstance(result, Graph)
#             assert len(result) == 0

#     def test_query_with_bnode(self, mock_oxigraph_url):
#         """Test query results containing blank nodes."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             bnode_response = Mock()
#             bnode_response.status_code = 200
#             bnode_response.headers = {"Content-Type": "application/sparql-results+json"}
#             bnode_response.text = """{
#                 "head": {"vars": ["s", "p", "o"]},
#                 "results": {
#                     "bindings": [{
#                         "s": {"type": "bnode", "value": "_:b1"},
#                         "p": {"type": "uri", "value": "http://example.org/property"},
#                         "o": {"type": "literal", "value": "test value"}
#                     }]
#                 }
#             }"""
#             bnode_response.raise_for_status = Mock()
#             mock_post.return_value = bnode_response

#             adapter = Oxigraph(mock_oxigraph_url)
#             result = adapter.query("SELECT ?s ?p ?o WHERE { ?s ?p ?o }")

#             rows = list(result)
#             assert len(rows) == 1
#             row = rows[0]

#             # Verify blank node is properly handled
#             from rdflib.term import BNode
#             assert isinstance(row.s, BNode)

#     def test_connection_test_with_custom_timeout(self):
#         """Test connection test with custom timeout."""
#         with patch('requests.get') as mock_get:
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()

#             adapter = Oxigraph("http://localhost:7878", timeout=30)

#             assert adapter.timeout == 30
#             # Verify connection test was called with correct timeout
#             mock_get.assert_called_once()
#             assert mock_get.call_args[1]['timeout'] == 30

#     def test_main_script_functionality(self):
#         """Test the main script functionality for basic operations."""
#         with patch('requests.get') as mock_get, \
#              patch('requests.post') as mock_post:

#             # Setup HTTP mocks
#             mock_get.return_value.status_code = 200
#             mock_get.return_value.raise_for_status = Mock()
#             mock_post.return_value.raise_for_status = Mock()

#             # Test with explicit parameters
#             adapter = Oxigraph(
#                 oxigraph_url="http://localhost:7878"
#             )

#             # Test that the main functionality can be called without errors
#             assert adapter.oxigraph_url == "http://localhost:7878"


class TestOxigraphIntegration:
    """Integration tests for Oxigraph (require running Oxigraph instance)."""

    @pytest.fixture
    def oxigraph_adapter(self):
        """Create an Oxigraph adapter for integration testing."""
        return Oxigraph(oxigraph_url="http://localhost:7878")

    @pytest.fixture
    def integration_graph(self):
        """Create a test graph for integration testing."""
        g = Graph()
        g.bind("test", "http://test.example.org/")

        subject = URIRef("http://test.example.org/integration_test")
        g.add((subject, RDF.type, URIRef("http://test.example.org/TestEntity")))
        g.add(
            (
                subject,
                URIRef("http://test.example.org/testProperty"),
                Literal("integration_value"),
            )
        )

        return g

    def generate_graph(
        self, num_triples: int, base_uri: str = "http://test.example.org/"
    ):
        """Create a large test graph for integration testing."""
        g = Graph()
        g.bind("test", base_uri)

        for i in range(num_triples):
            subject = URIRef(f"{base_uri}/entity{i}")
            g.add((subject, RDF.type, URIRef(f"{base_uri}/TestEntity")))

        return g

    def test_large_graph_insert(self, oxigraph_adapter):
        """Test inserting a large graph into Oxigraph."""
        try:
            base_uri = f"http://test.example.org/{uuid.uuid4()}"
            large_graph = self.generate_graph(100001, base_uri)
            oxigraph_adapter.insert(large_graph)

            # Verify the data was inserted
            count_query = (
                f"SELECT (COUNT(*) as ?count) WHERE {{ ?s a <{base_uri}/TestEntity> }}"
            )
            result = list(oxigraph_adapter.query(count_query))
            assert len(result) == 1

            print(result)

            # Access count using Variable object
            from rdflib.term import Variable

            count_var = Variable("count")
            assert int(result[0][count_var]) == 100001

            # Cleanup
            oxigraph_adapter.remove(large_graph)
        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping large graph test")

    def test_insert_and_query_workflow(self, oxigraph_adapter, integration_graph):
        """Test basic insert and query workflow."""
        try:
            # Insert test data
            oxigraph_adapter.insert(integration_graph)

            # Query for the data
            query = """
            SELECT ?s ?p ?o WHERE {
                ?s <http://test.example.org/testProperty> ?o
            }
            """
            results = list(oxigraph_adapter.query(query))
            assert len(results) == 1
            assert str(results[0].o) == "integration_value"

            # Cleanup
            oxigraph_adapter.remove(integration_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping integration test")

    def test_construct_query(self, oxigraph_adapter, integration_graph):
        """Test CONSTRUCT query returns a graph."""
        try:
            oxigraph_adapter.insert(integration_graph)

            query = """
            CONSTRUCT { ?s ?p ?o }
            WHERE { 
                ?s <http://test.example.org/testProperty> ?o .
                ?s ?p ?o
            }
            """
            result = oxigraph_adapter.query(query)

            assert isinstance(result, Graph)
            assert len(result) > 0

            # Cleanup
            oxigraph_adapter.remove(integration_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping construct test")

    def test_ask_query(self, oxigraph_adapter, integration_graph):
        """Test ASK query."""
        try:
            oxigraph_adapter.insert(integration_graph)

            # Ask if data exists
            ask_query = """
            ASK WHERE {
                ?s <http://test.example.org/testProperty> "integration_value"
            }
            """
            result = list(oxigraph_adapter.query(ask_query))
            assert result is not None

            # Cleanup
            oxigraph_adapter.remove(integration_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping ASK test")

    def test_get_subject_graph(self, oxigraph_adapter, integration_graph):
        """Test retrieving all triples for a specific subject."""
        try:
            oxigraph_adapter.insert(integration_graph)

            subject = URIRef("http://test.example.org/integration_test")
            subject_graph = oxigraph_adapter.get_subject_graph(subject)

            assert isinstance(subject_graph, Graph)
            assert len(subject_graph) == 2  # type and testProperty

            # Cleanup
            oxigraph_adapter.remove(integration_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping subject graph test")

    def test_filter_query(self, oxigraph_adapter):
        """Test SPARQL FILTER clause."""
        try:
            # Create test data with numbers - use unique URI path
            g = Graph()
            g.bind("test", "http://test.example.org/filter/")

            for i in range(10):
                subject = URIRef(f"http://test.example.org/filter/number{i}")
                g.add(
                    (subject, RDF.type, URIRef("http://test.example.org/filter/Number"))
                )
                g.add(
                    (
                        subject,
                        URIRef("http://test.example.org/filter/value"),
                        Literal(i),
                    )
                )

            oxigraph_adapter.insert(g)

            # Query with FILTER - use specific URI
            query = """
            SELECT ?s ?value WHERE {
                ?s <http://test.example.org/filter/value> ?value .
                FILTER(?value > 5)
            }
            ORDER BY ?value
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 4  # values 6, 7, 8, 9
            from rdflib.term import Variable

            value_var = Variable("value")
            assert int(results[0][value_var]) == 6
            assert int(results[-1][value_var]) == 9

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping FILTER test")

    def test_optional_clause(self, oxigraph_adapter):
        """Test SPARQL OPTIONAL clause."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            # Create entities with and without optional properties
            for i in range(5):
                subject = URIRef(f"http://test.example.org/entity{i}")
                g.add((subject, RDF.type, URIRef("http://test.example.org/Entity")))
                g.add((subject, URIRef("http://test.example.org/id"), Literal(i)))

                # Only add optional property for even numbers
                if i % 2 == 0:
                    g.add(
                        (
                            subject,
                            URIRef("http://test.example.org/optional"),
                            Literal(f"value{i}"),
                        )
                    )

            oxigraph_adapter.insert(g)

            query = """
            SELECT ?s ?id ?optional WHERE {
                ?s <http://test.example.org/id> ?id .
                OPTIONAL { ?s <http://test.example.org/optional> ?optional }
            }
            ORDER BY ?id
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 5
            # Check that optional values are present for even IDs
            for row in results:
                if int(row.id) % 2 == 0:
                    assert row.optional is not None
                else:
                    assert row.optional is None

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping OPTIONAL test")

    def test_union_clause(self, oxigraph_adapter):
        """Test SPARQL UNION clause."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            # Create different types of entities
            person = URIRef("http://test.example.org/person1")
            g.add((person, RDF.type, URIRef("http://test.example.org/Person")))
            g.add((person, URIRef("http://test.example.org/name"), Literal("Alice")))

            org = URIRef("http://test.example.org/org1")
            g.add((org, RDF.type, URIRef("http://test.example.org/Organization")))
            g.add((org, URIRef("http://test.example.org/title"), Literal("ACME Corp")))

            oxigraph_adapter.insert(g)

            query = """
            SELECT ?entity ?label WHERE {
                {
                    ?entity a <http://test.example.org/Person> .
                    ?entity <http://test.example.org/name> ?label .
                }
                UNION
                {
                    ?entity a <http://test.example.org/Organization> .
                    ?entity <http://test.example.org/title> ?label .
                }
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 2
            labels = [str(row.label) for row in results]
            assert "Alice" in labels
            assert "ACME Corp" in labels

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping UNION test")

    def test_aggregation_functions(self, oxigraph_adapter):
        """Test SPARQL aggregation functions (COUNT, SUM, AVG, MIN, MAX)."""
        try:
            from rdflib.term import Variable

            g = Graph()
            g.bind("test", "http://test.example.org/agg/")

            values = [10, 20, 30, 40, 50]
            for i, val in enumerate(values):
                subject = URIRef(f"http://test.example.org/agg/item{i}")
                g.add(
                    (subject, URIRef("http://test.example.org/agg/value"), Literal(val))
                )

            oxigraph_adapter.insert(g)

            # Test COUNT, SUM, AVG, MIN, MAX
            query = """
            SELECT 
                (COUNT(?value) as ?count)
                (SUM(?value) as ?sum)
                (AVG(?value) as ?avg)
                (MIN(?value) as ?min)
                (MAX(?value) as ?max)
            WHERE {
                ?s <http://test.example.org/agg/value> ?value .
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 1
            row = results[0]
            count_var = Variable("count")
            sum_var = Variable("sum")
            avg_var = Variable("avg")
            min_var = Variable("min")
            max_var = Variable("max")

            assert int(row[count_var]) == 5
            assert int(row[sum_var]) == 150
            assert int(row[avg_var]) == 30
            assert int(row[min_var]) == 10
            assert int(row[max_var]) == 50

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping aggregation test")

    def test_group_by_having(self, oxigraph_adapter):
        """Test SPARQL GROUP BY and HAVING clauses."""
        try:
            from rdflib.term import Variable

            g = Graph()
            g.bind("test", "http://test.example.org/groupby/")

            # Create categories with items
            categories = ["A", "B", "C"]
            for cat in categories:
                for i in range(3 if cat == "A" else 2 if cat == "B" else 1):
                    subject = URIRef(f"http://test.example.org/groupby/{cat}/item{i}")
                    g.add(
                        (
                            subject,
                            URIRef("http://test.example.org/groupby/category"),
                            Literal(cat),
                        )
                    )

            oxigraph_adapter.insert(g)

            query = """
            SELECT ?category (COUNT(?item) as ?count)
            WHERE {
                ?item <http://test.example.org/groupby/category> ?category .
            }
            GROUP BY ?category
            HAVING (COUNT(?item) > 1)
            ORDER BY DESC(?count)
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 2  # Only A and B have more than 1 item
            category_var = Variable("category")
            count_var = Variable("count")

            assert str(results[0][category_var]) == "A"
            assert int(results[0][count_var]) == 3
            assert str(results[1][category_var]) == "B"
            assert int(results[1][count_var]) == 2

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping GROUP BY test")

    def test_blank_nodes_are_filtered(self, oxigraph_adapter):
        """Test that blank nodes are filtered out during insert."""
        try:
            domain = uuid.uuid4()
            g = Graph()

            # Create a structure with blank nodes
            person = URIRef(f"http://{domain}/bnode/person1")
            address = BNode()

            g.add((person, RDF.type, URIRef(f"http://{domain}/bnode/Person")))
            g.add((person, URIRef(f"http://{domain}/bnode/hasAddress"), address))
            g.add(
                (
                    address,
                    URIRef(f"http://{domain}/bnode/street"),
                    Literal("123 Main St"),
                )
            )

            # Insert graph with blank nodes
            oxigraph_adapter.insert(g)

            # Query should only find the person type, not the address relationship
            query = f"""
            SELECT ?s ?p ?o WHERE {{
                ?s ?p ?o .
                FILTER(STRSTARTS(STR(?s), "http://{domain}/"))
            }}
            """
            results = list(oxigraph_adapter.query(query))

            # Should only have 1 triple (person rdf:type Person)
            # The blank node triples should have been filtered out
            assert len(results) == 1

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running")

    def test_special_characters_in_literals(self, oxigraph_adapter):
        """Test handling of special characters in literals."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            subject = URIRef("http://test.example.org/special")

            # Test various special characters
            special_strings = [
                'String with "quotes"',
                "String with 'apostrophes'",
                "String with\nnewlines",
                "String with\ttabs",
                "Unicode: 你好世界 🌍",
                "Emoji: 😀 👍 🎉",
                "Special chars: @#$%^&*()",
                "Backslash: \\test\\path",
            ]

            for i, text in enumerate(special_strings):
                g.add(
                    (subject, URIRef(f"http://test.example.org/prop{i}"), Literal(text))
                )

            oxigraph_adapter.insert(g)

            # Query back and verify
            query = """
            SELECT ?p ?o WHERE {
                <http://test.example.org/special> ?p ?o .
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == len(special_strings)

            retrieved_values = [str(row.o) for row in results]
            for original in special_strings:
                assert original in retrieved_values

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping special characters test")

    def test_different_datatypes(self, oxigraph_adapter):
        """Test handling of different RDF datatypes."""
        try:
            from rdflib.namespace import XSD

            g = Graph()
            g.bind("test", "http://test.example.org/")

            subject = URIRef("http://test.example.org/datatypes")

            # Add different datatype literals
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/integer"),
                    Literal(42, datatype=XSD.integer),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/float"),
                    Literal(3.14, datatype=XSD.float),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/boolean"),
                    Literal(True, datatype=XSD.boolean),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/date"),
                    Literal("2024-01-01", datatype=XSD.date),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/string"),
                    Literal("test string", datatype=XSD.string),
                )
            )

            oxigraph_adapter.insert(g)

            # Query and verify datatypes are preserved
            query = """
            SELECT ?p ?o WHERE {
                <http://test.example.org/datatypes> ?p ?o .
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 5

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping datatypes test")

    def test_language_tags(self, oxigraph_adapter):
        """Test handling of language-tagged literals."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            subject = URIRef("http://test.example.org/multilingual")

            # Add same property with different language tags
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/label"),
                    Literal("Hello", lang="en"),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/label"),
                    Literal("Bonjour", lang="fr"),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/label"),
                    Literal("Hola", lang="es"),
                )
            )
            g.add(
                (
                    subject,
                    URIRef("http://test.example.org/label"),
                    Literal("你好", lang="zh"),
                )
            )

            oxigraph_adapter.insert(g)

            # Query with language filter
            query = """
            SELECT ?label WHERE {
                <http://test.example.org/multilingual> <http://test.example.org/label> ?label .
                FILTER(LANG(?label) = "fr")
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 1
            assert str(results[0].label) == "Bonjour"

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping language tags test")

    def test_property_paths(self, oxigraph_adapter):
        """Test SPARQL property paths."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            # Create a chain: A -> B -> C -> D
            for i in range(3):
                subject = URIRef(f"http://test.example.org/node{i}")
                obj = URIRef(f"http://test.example.org/node{i + 1}")
                g.add((subject, URIRef("http://test.example.org/next"), obj))

            oxigraph_adapter.insert(g)

            # Test property path with +
            query = """
            SELECT ?end WHERE {
                <http://test.example.org/node0> <http://test.example.org/next>+ ?end .
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 3  # node1, node2, node3

            # Cleanup
            oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping property paths test")

    def test_describe_query(self, oxigraph_adapter, integration_graph):
        """Test DESCRIBE query."""
        try:
            oxigraph_adapter.insert(integration_graph)

            query = """
            DESCRIBE <http://test.example.org/integration_test>
            """
            result = oxigraph_adapter.query(query)

            assert isinstance(result, Graph)
            assert len(result) > 0

            # Cleanup
            oxigraph_adapter.remove(integration_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping DESCRIBE test")

    def test_empty_graph_operations(self, oxigraph_adapter):
        """Test operations with empty graphs."""
        try:
            empty_graph = Graph()

            # Insert empty graph should not fail
            oxigraph_adapter.insert(empty_graph)

            # Remove empty graph should not fail
            oxigraph_adapter.remove(empty_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping empty graph test")

    def test_update_operations(self, oxigraph_adapter):
        """Test SPARQL UPDATE operations."""
        try:
            g = Graph()
            g.bind("test", "http://test.example.org/")

            subject = URIRef("http://test.example.org/updatetest")
            g.add(
                (subject, URIRef("http://test.example.org/value"), Literal("initial"))
            )

            oxigraph_adapter.insert(g)

            # Update using SPARQL DELETE/INSERT
            update_query = """
            DELETE { <http://test.example.org/updatetest> <http://test.example.org/value> ?old }
            INSERT { <http://test.example.org/updatetest> <http://test.example.org/value> "updated" }
            WHERE { <http://test.example.org/updatetest> <http://test.example.org/value> ?old }
            """
            oxigraph_adapter.query(update_query)

            # Query to verify update
            query = """
            SELECT ?value WHERE {
                <http://test.example.org/updatetest> <http://test.example.org/value> ?value .
            }
            """
            results = list(oxigraph_adapter.query(query))

            assert len(results) == 1
            assert str(results[0].value) == "updated"

            # Cleanup - remove all
            cleanup_graph = Graph()
            cleanup_graph.add(
                (subject, URIRef("http://test.example.org/value"), Literal("updated"))
            )
            oxigraph_adapter.remove(cleanup_graph)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping UPDATE test")

    def test_batch_insertions(self, oxigraph_adapter):
        """Test multiple batch insertions."""
        try:
            from rdflib.term import Variable

            graphs = []

            # Create multiple graphs - use unique URI
            for batch in range(5):
                g = Graph()
                g.bind("test", "http://test.example.org/batches/")

                for i in range(100):
                    subject = URIRef(
                        f"http://test.example.org/batches/batch{batch}/item{i}"
                    )
                    g.add(
                        (
                            subject,
                            RDF.type,
                            URIRef("http://test.example.org/batches/Item"),
                        )
                    )
                    g.add(
                        (
                            subject,
                            URIRef("http://test.example.org/batches/batch"),
                            Literal(batch),
                        )
                    )

                graphs.append(g)
                oxigraph_adapter.insert(g)

            # Verify total count
            count_query = """
            SELECT (COUNT(*) as ?count) WHERE {
                ?s a <http://test.example.org/batches/Item> .
            }
            """
            results = list(oxigraph_adapter.query(count_query))
            count_var = Variable("count")
            assert int(results[0][count_var]) == 500  # 5 batches * 100 items

            # Cleanup
            for g in graphs:
                oxigraph_adapter.remove(g)

        except requests.exceptions.ConnectionError:
            pytest.skip("Oxigraph is not running - skipping batch insertions test")

    # @pytest.mark.integration
    # def test_full_workflow(self, oxigraph_adapter, integration_graph):
    #     """Test complete workflow: insert, query, get subject, remove."""
    #     try:
    #         # Insert test data
    #         oxigraph_adapter.insert(integration_graph)

    #         # Query for the data
    #         query = """
    #         SELECT ?s ?p ?o WHERE {
    #             ?s <http://test.example.org/testProperty> ?o
    #         }
    #         """
    #         results = list(oxigraph_adapter.query(query))
    #         assert len(results) == 1

    #         # Get subject graph
    #         subject = URIRef("http://test.example.org/integration_test")
    #         subject_graph = oxigraph_adapter.get_subject_graph(subject)
    #         assert len(subject_graph) == 2  # type and testProperty

    #         # Remove the data
    #         oxigraph_adapter.remove(integration_graph)

    #         # Verify removal
    #         results_after_remove = list(oxigraph_adapter.query(query))
    #         assert len(results_after_remove) == 0

    #     except requests.exceptions.ConnectionError:
    #         pytest.skip("Oxigraph is not running - skipping integration test")

    # @pytest.mark.integration
    # def test_large_dataset_performance(self, oxigraph_adapter):
    #     """Test performance with larger datasets."""
    #     try:
    #         # Create a larger test dataset
    #         large_graph = Graph()
    #         large_graph.bind("perf", "http://performance.test.org/")

    #         # Add 1000 triples
    #         for i in range(1000):
    #             subject = URIRef(f"http://performance.test.org/entity{i}")
    #             large_graph.add((subject, RDF.type, URIRef("http://performance.test.org/Entity")))
    #             large_graph.add((subject, URIRef("http://performance.test.org/id"), Literal(i)))

    #         # Measure insert time
    #         start_time = time.time()
    #         oxigraph_adapter.insert(large_graph)
    #         insert_time = time.time() - start_time

    #         # Insert should complete in reasonable time (Oxigraph is fast!)
    #         assert insert_time < 5.0, f"Insert took too long: {insert_time}s"

    #         # Test query performance
    #         start_time = time.time()
    #         results = list(oxigraph_adapter.query("SELECT ?s WHERE { ?s a <http://performance.test.org/Entity> }"))
    #         query_time = time.time() - start_time

    #         assert len(results) == 1000
    #         assert query_time < 2.0, f"Query took too long: {query_time}s"

    #         # Test COUNT query performance
    #         start_time = time.time()
    #         count_results = list(oxigraph_adapter.query("SELECT (COUNT(*) as ?count) WHERE { ?s a <http://performance.test.org/Entity> }"))
    #         count_time = time.time() - start_time

    #         assert len(count_results) == 1
    #         assert int(count_results[0].count) == 1000
    #         assert count_time < 1.0, f"Count query took too long: {count_time}s"

    #         # Clean up
    #         oxigraph_adapter.remove(large_graph)

    #     except requests.exceptions.ConnectionError:
    #         pytest.skip("Oxigraph is not running - skipping performance test")

    # @pytest.mark.integration
    # def test_concurrent_operations(self, oxigraph_adapter):
    #     """Test concurrent insert and query operations."""
    #     try:
    #         import concurrent.futures

    #         def insert_data(thread_id):
    #             """Insert data in a separate thread."""
    #             graph = Graph()
    #             for i in range(100):
    #                 subject = URIRef(f"http://concurrent.test.org/thread{thread_id}/entity{i}")
    #                 graph.add((subject, RDF.type, URIRef("http://concurrent.test.org/Entity")))
    #                 graph.add((subject, URIRef("http://concurrent.test.org/threadId"), Literal(thread_id)))
    #             oxigraph_adapter.insert(graph)
    #             return thread_id

    #         # Run multiple insert operations concurrently
    #         with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
    #             futures = [executor.submit(insert_data, i) for i in range(3)]
    #             results = [future.result() for future in concurrent.futures.as_completed(futures)]

    #         assert len(results) == 3

    #         # Verify all data was inserted correctly
    #         total_results = list(oxigraph_adapter.query(
    #             "SELECT ?s WHERE { ?s a <http://concurrent.test.org/Entity> }"
    #         ))
    #         assert len(total_results) == 300  # 3 threads × 100 entities each

    #         # Clean up
    #         cleanup_graph = Graph()
    #         for result in total_results:
    #             subject_graph = oxigraph_adapter.get_subject_graph(result.s)
    #             cleanup_graph += subject_graph
    #         oxigraph_adapter.remove(cleanup_graph)

    #     except requests.exceptions.ConnectionError:
    #         pytest.skip("Oxigraph is not running - skipping concurrent test")

    # @pytest.mark.integration
    # def test_complex_sparql_queries(self, oxigraph_adapter):
    #     """Test complex SPARQL queries with various features."""
    #     try:
    #         # Setup test data
    #         test_graph = Graph()
    #         test_graph.bind("test", "http://complex.test.org/")

    #         # Create hierarchical data
    #         for i in range(5):
    #             person = URIRef(f"http://complex.test.org/person{i}")
    #             test_graph.add((person, RDF.type, URIRef("http://complex.test.org/Person")))
    #             test_graph.add((person, URIRef("http://complex.test.org/name"), Literal(f"Person {i}")))
    #             test_graph.add((person, URIRef("http://complex.test.org/age"), Literal(20 + i)))

    #             if i > 0:
    #                 friend = URIRef(f"http://complex.test.org/person{i-1}")
    #                 test_graph.add((person, URIRef("http://complex.test.org/knows"), friend))

    #         oxigraph_adapter.insert(test_graph)

    #         # Test FILTER query
    #         filter_results = list(oxigraph_adapter.query("""
    #             PREFIX test: <http://complex.test.org/>
    #             SELECT ?person ?age WHERE {
    #                 ?person a test:Person ;
    #                         test:age ?age .
    #                 FILTER(?age > 22)
    #             }
    #             ORDER BY ?age
    #         """))

    #         assert len(filter_results) == 2  # persons with age > 22

    #         # Test OPTIONAL query
    #         optional_results = list(oxigraph_adapter.query("""
    #             PREFIX test: <http://complex.test.org/>
    #             SELECT ?person ?name ?friend WHERE {
    #                 ?person a test:Person ;
    #                         test:name ?name .
    #                 OPTIONAL { ?person test:knows ?friend }
    #             }
    #             ORDER BY ?name
    #         """))

    #         assert len(optional_results) == 5  # All persons

    #         # Test GROUP BY and COUNT
    #         group_results = list(oxigraph_adapter.query("""
    #             PREFIX test: <http://complex.test.org/>
    #             SELECT (COUNT(?friend) as ?friendCount) WHERE {
    #                 ?person a test:Person .
    #                 OPTIONAL { ?person test:knows ?friend }
    #             }
    #         """))

    #         assert len(group_results) == 1
    #         assert int(group_results[0].friendCount) == 4  # Total friendship connections

    #         # Clean up
    #         oxigraph_adapter.remove(test_graph)

    #     except requests.exceptions.ConnectionError:
    #         pytest.skip("Oxigraph is not running - skipping complex query test")


if __name__ == "__main__":
    # Run basic tests
    pytest.main([__file__, "-v"])



import os
from queue import Queue
from typing import Optional

from naas_abi_core.services.object_storage.ObjectStoragePort import (
    Exceptions,
    IObjectStorageAdapter,
)


class ObjectStorageSecondaryAdapterFS(IObjectStorageAdapter):
    def __init__(self, base_path: str):
        self.base_path = base_path
        self.__create_path(base_path)

    def __create_path(self, prefix: str) -> None:
        os.makedirs(os.path.join(self.base_path, prefix), exist_ok=True)

    def __path_exists(self, prefix: str, key: str | None = None) -> bool:
        if key is None:
            exists = os.path.exists(os.path.join(self.base_path, prefix))
        else:
            exists = os.path.exists(os.path.join(self.base_path, prefix, key))

        if not exists:
            raise Exceptions.ObjectNotFound(f"Object {prefix}/{key} not found")

        return exists

    def get_object(self, prefix: str, key: str) -> bytes:
        self.__path_exists(prefix, key)

        with open(os.path.join(self.base_path, prefix, key), "rb") as f:
            return f.read()

    def put_object(self, prefix: str, key: str, content: bytes) -> None:
        self.__create_path(prefix)

        with open(os.path.join(self.base_path, prefix, key), "wb") as f:
            f.write(content)

    def delete_object(self, prefix: str, key: str) -> None:
        self.__path_exists(prefix, key)

        os.remove(os.path.join(self.base_path, prefix, key))

    def list_objects(self, prefix: str, queue: Optional[Queue] = None) -> list[str]:
        self.__path_exists(prefix)
        return [
            os.path.join(prefix, f)
            for f in os.listdir(os.path.join(self.base_path, prefix))
        ]



import logging
import os
from typing import Optional

from .IVectorStorePort import IVectorStorePort
from .VectorStoreService import VectorStoreService

logger = logging.getLogger(__name__)


class VectorStoreFactory:
    _instance: Optional[VectorStoreService] = None

    @classmethod
    def create_adapter(cls) -> IVectorStorePort:
        adapter_type = os.getenv("VECTOR_STORE_ADAPTER", "qdrant").lower()

        if adapter_type == "qdrant":
            # Lazy import to avoid loading qdrant_client at module import time
            from .adapters.QdrantAdapter import QdrantAdapter

            host = os.getenv("QDRANT_HOST", "localhost")
            port = int(os.getenv("QDRANT_PORT", "6333"))
            api_key = os.getenv("QDRANT_API_KEY")
            https = os.getenv("QDRANT_HTTPS", "false").lower() == "true"
            timeout = int(os.getenv("QDRANT_TIMEOUT", "30"))

            logger.info(f"Creating Qdrant adapter (host={host}, port={port})")
            return QdrantAdapter(
                host=host, port=port, api_key=api_key, https=https, timeout=timeout
            )
        else:
            raise ValueError(f"Unknown vector store adapter: {adapter_type}")

    @classmethod
    def get_service(cls) -> VectorStoreService:
        if cls._instance is None:
            adapter = cls.create_adapter()
            cls._instance = VectorStoreService(adapter)
            logger.info("VectorStoreService singleton created")
        return cls._instance

    @classmethod
    def reset(cls) -> None:
        if cls._instance:
            logger.info("Resetting VectorStoreService singleton")
        cls._instance = None



import os
from typing import Any, Dict

from dotenv import dotenv_values, find_dotenv, set_key
from naas_abi_core.services.secret.SecretPorts import ISecretAdapter
from naas_abi_core.utils.Logger import logger


class DotenvSecretSecondaryAdaptor(ISecretAdapter):
    def __init__(self):
        logger.debug(f"dotenv is using the file: {find_dotenv()}")

        self.secrets = dotenv_values()

    def get(self, key: str, default: Any = None) -> str:
        return self.secrets.get(key, os.environ.get(key, default))

    def set(self, key: str, value: str):
        os.environ[key] = value
        set_key(".env", key, value)

    def remove(self, key: str):
        os.environ.pop(key, None)

    def list(self) -> Dict[str, str | None]:
        secrets: Dict[str, str | None] = {}
        for key in self.secrets.keys():
            secrets[str(key)] = str(self.secrets.get(key))
        return secrets



from typing import Dict, List, Optional

import rdflib
from rdflib import DCTERMS, OWL, RDF, RDFS, XSD, Graph, URIRef, query

from naas_abi_core import logger
from naas_abi_core.services.triple_store.TripleStorePorts import ITripleStoreService
from naas_abi_core.utils.Graph import ABI, BFO, CCO, TEST


class SPARQLUtils:
    __triple_store_service: ITripleStoreService

    def __init__(self, triple_store_service: ITripleStoreService):
        self.__triple_store_service = triple_store_service

    @property
    def triple_store_service(self) -> ITripleStoreService:
        return self.__triple_store_service

    def results_to_list(self, results: rdflib.query.Result) -> Optional[List[Dict]]:
        """
        Transform SPARQL query results to a list of dictionaries using the labels as keys.

        Args:
            results (query.Result): The SPARQL query results to transform

        Returns:
            Optional[List[Dict]]: List of dictionaries with query results, or None if no results
        """
        data = []
        for row in results:
            assert isinstance(row, query.ResultRow)
            logger.debug(f"==> Row: {row}")
            data_dict = {}
            for key in row.labels:
                data_dict[key] = str(row[key]) if row[key] else None
            data.append(data_dict)
        return data if len(data) > 0 else None

    def get_class_uri_from_individual_uri(
        self,
        uri: str | URIRef,
    ) -> Optional[str]:
        """
        Get the class URI for a given individual URI from the triple store.

        Args:
            uri (str | URIRef): The individual URI to look up

        Returns:
            Optional[str]: The class URI if found, None otherwise
        """
        sparql_query = f"""
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            SELECT ?type
            WHERE {{
                <{uri}> rdf:type ?type .
                FILTER(?type != owl:NamedIndividual)
            }}
            LIMIT 1
        """

        try:
            results = self.triple_store_service.query(sparql_query)
            for row in results:
                assert isinstance(row, query.ResultRow)
                return URIRef(str(row.type))
            return None
        except Exception as e:
            logger.error(f"Error getting class URI for {uri}: {e}")
            return None

    def get_rdfs_label_from_individual_uri(
        self,
        uri: str | URIRef,
    ) -> Optional[str]:
        """
        Get the RDFS label for a given individual URI from the triple store.

        Args:
            uri (str | URIRef): The individual URI to look up

        Returns:
            Optional[str]: The RDFS label if found, None otherwise
        """
        sparql_query = f"""
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            SELECT ?label
            WHERE {{
                <{uri}> rdfs:label ?label .
            }}
            LIMIT 1
        """

        try:
            results = self.triple_store_service.query(sparql_query)
            for row in results:
                assert isinstance(row, query.ResultRow)
                return str(row.label)
            return None
        except Exception as e:
            logger.error(f"Error getting label for {uri}: {e}")
            return None

    def get_identifier(
        self,
        identifier: str,
        type: URIRef = URIRef("http://ontology.naas.ai/abi/unique_id"),
        graph: Graph = Graph(),
    ) -> Optional[URIRef]:
        """
        Get the URI for a given identifier from the triple store or provided graph.

        Args:
            identifier (str): The identifier string to look up
            type (URIRef, optional): The predicate type to use for the lookup.
                                    Defaults to "http://ontology.naas.ai/abi/unique_id"
            graph (Graph, optional): Optional RDFlib Graph to query instead of triple store.
                                Defaults to empty Graph.

        Returns:
            Optional[URIRef]: The URI if found, None otherwise
        """
        sparql_query = f"""
            SELECT ?s
            WHERE {{
                ?s <{str(type)}> "{identifier}" .
            }}
            LIMIT 1
        """
        try:
            if len(graph) > 0:
                results = graph.query(sparql_query)
            else:
                results = self.query(sparql_query)  # type: ignore

            for row in results:
                assert isinstance(row, query.ResultRow)
                # Use existing URI if found
                return URIRef(str(row.s))
        except Exception as e:
            logger.error(f"Error getting identifier for {identifier}: {e}")
            return None
        return None

    def get_identifiers(
        self,
        property_uri: URIRef = URIRef("http://ontology.naas.ai/abi/unique_id"),
        class_uri: Optional[URIRef] = None,
    ) -> dict[str, URIRef]:
        """
        Get a mapping of all identifiers to their URIs from the triple store.

        Args:
            property_uri (URIRef, optional): The predicate URI to use for the lookup.
                                        Defaults to "http://ontology.naas.ai/abi/unique_id"
            class_uri (URIRef, optional): Optional class URI to filter results.
                                        Only return identifiers for instances of this class.
                                        Defaults to None.

        Returns:
            dict[str, URIRef]: Dictionary mapping identifiers to their URIs
        """
        sparql_query = f"""
            SELECT ?s ?id
            WHERE {{
                ?s <{str(property_uri)}> ?id .
                {f"?s a <{str(class_uri)}> ." if class_uri else ""}
            }}
        """
        try:
            results = self.query(sparql_query)  # type: ignore

            id_map = {}
            for row in results:
                assert isinstance(row, query.ResultRow)
                id_map[str(row.id)] = URIRef(str(row.s))
            return id_map
        except Exception as e:
            logger.error(f"Error getting identifiers map: {e}")
            return {}

    def get_subject_graph(self, uri: str | URIRef, depth: int = 1) -> Graph:
        """
        Get a graph for a given URI with a specified depth of relationships.
        This recursively follows relationships to build a more detailed subgraph.
        The resulting graph includes all triples where the given URI is the subject,
        and optionally follows object URIs to include their relationships up to the specified depth.

        Args:
            uri (str | URIRef): The URI to build the graph around
            depth (int): How many levels deep to traverse relationships. A depth of 0 returns an empty graph,
                        1 returns direct relationships, 2 includes relationships of related objects, etc.
                        Defaults to 1.

        Returns:
            Graph: RDFlib Graph containing all triples within the specified depth, with standard namespace
                prefixes bound (rdfs, rdf, owl, xsd, dcterms, abi, bfo, cco, test)
        """
        if depth <= 0:
            return Graph()

        # Build the CONSTRUCT query dynamically based on depth
        construct_clauses = []
        where_clauses = []

        # Add patterns for each depth level
        for i in range(depth):
            if i == 0:
                construct_clauses.append(f"<{str(uri)}> ?p{i} ?o{i} .")
                where_clauses.append(f"<{str(uri)}> ?p{i} ?o{i} .")
            else:
                construct_clauses.append(f"?o{i - 1} ?p{i} ?o{i} .")
                where_clauses.append(
                    f"OPTIONAL {{ ?o{i - 1} ?p{i} ?o{i} . FILTER(isURI(?o{i - 1})) }}"
                )

        sparql_query = f"""
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
            PREFIX dcterms: <http://purl.org/dc/terms/>
            PREFIX abi: <http://ontology.naas.ai/abi/>
            PREFIX bfo: <http://purl.obolibrary.org/obo/>
            PREFIX cco: <https://www.commoncoreontologies.org/>
            CONSTRUCT {{ 
                {" ".join(construct_clauses)}
            }}
            WHERE {{
                {" ".join(where_clauses)}
            }}
        """
        try:
            results = self.triple_store_service.query(sparql_query)
        except Exception as e:
            logger.error(f"Error getting subject graph for {uri}: {e}")
            return Graph()

        graph = Graph()
        graph.bind("rdfs", RDFS)
        graph.bind("rdf", RDF)
        graph.bind("owl", OWL)
        graph.bind("xsd", XSD)
        graph.bind("dcterms", DCTERMS)
        graph.bind("abi", ABI)
        graph.bind("bfo", BFO)
        graph.bind("cco", CCO)
        graph.bind("test", TEST)
        for triple in results:
            # CONSTRUCT queries return triples directly, no need for ResultRow handling
            graph.add(triple)  # type: ignore

        return graph



from datetime import date, datetime
from urllib.parse import quote

import pytz
from rdflib import Graph as rdfgraph
from rdflib import Literal, Namespace, URIRef
from rdflib.namespace import OWL, RDF, RDFS, SKOS

from naas_abi_core import logger

BFO = Namespace("http://purl.obolibrary.org/obo/")
ABI = Namespace("http://ontology.naas.ai/abi/")
TEST = Namespace("http://ontology.naas.ai/test/")
TIME = Namespace("http://www.w3.org/2006/time#")
XSD = Namespace("http://www.w3.org/2001/XMLSchema#")
CCO = Namespace("https://www.commoncoreontologies.org/")
DCTERMS = Namespace("http://purl.org/dc/terms/")
URI_REGEX = r"http:\/\/ontology\.naas\.ai\/.+\/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"


class ABIGraph(rdfgraph):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.bind("bfo", BFO)
        self.bind("skos", SKOS)
        self.bind("abi", ABI)
        self.bind("cco", CCO)
        self.bind("xsd", XSD)
        self.bind("time", TIME)

    def add_data_properties(self, uri: URIRef, lang="en", **data_properties):
        for x in data_properties:
            value = data_properties.get(x)
            if value is not None:
                if type(value) is str:
                    self.add((uri, ABI[x], Literal(value.strip(), lang=lang)))
                elif type(value) in [int, float]:
                    self.add((uri, ABI[x], Literal(value, datatype=XSD.integer)))
                elif isinstance(value, datetime):
                    if value.tzinfo is None:
                        value = value.replace(
                            tzinfo=pytz.utc
                        )  # Apply UTC timezone if none
                    self.add(
                        (
                            uri,
                            ABI[x],
                            Literal(
                                value.strftime("%Y-%m-%dT%H:%M:%S%z"),
                                datatype=XSD.dateTime,
                            ),
                        )
                    )
                elif isinstance(value, date):
                    self.add(
                        (
                            uri,
                            ABI[x],
                            Literal(value.strftime("%Y-%m-%d"), datatype=XSD.date),
                        )
                    )
        self.add(
            (
                uri,
                DCTERMS.modified,
                Literal(
                    str(datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")),
                    datatype=XSD.dateTime,
                ),
            )
        )

    # Add OWL NamedIndividual to Graph
    def add_individual(
        self,
        uri: URIRef,
        label,
        is_a,
        lang="en",
        skip_if_exists=True,
        **data_properties,
    ) -> URIRef:
        if (uri, RDF.type, is_a) in self and skip_if_exists:
            logger.debug(f"🟡 '{label}' ({str(uri)}) already exists in ontology.")
        else:
            # Add NamedIndividual to ontology
            self.add((uri, RDF.type, OWL.NamedIndividual))
            self.add((uri, RDF.type, URIRef(is_a)))
            self.add((uri, RDFS.label, Literal(str(label), lang=lang)))
            self.add(
                (
                    uri,
                    DCTERMS.created,
                    Literal(
                        str(datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")),
                        datatype=XSD.dateTime,
                    ),
                )
            )
            logger.debug(f"🟢 '{label}' ({str(uri)}) successfully added to ontology.")

        # Add data properties to NamedIndividual
        self.add_data_properties(uri, lang, **data_properties)
        return uri

    def add_individual_to_prefix(
        self,
        prefix: Namespace,
        uid: str,
        label: str,
        is_a: URIRef,
        lang="en",
        skip_if_exists=True,
        **data_properties,
    ) -> URIRef:
        uid = str(uid).split(":")[-1]
        type_name = str(is_a).split("/")[-1]
        uri = URIRef(quote(f"{str(prefix)}{type_name}#{uid}", safe=":/#"))
        return self.add_individual(
            uri, label, is_a, lang, skip_if_exists, **data_properties
        )

    def add_process(
        self,
        prefix: Namespace,
        uid: str,
        label: str,
        is_a: URIRef,
        lang="en",
        participants=[],
        participants_oprop=BFO.BFO_0000057,
        participants_oprop_inverse=BFO.BFO_0000056,
        realizes=[],
        realizes_oprop=BFO.BFO_0000055,
        realizes_oprop_inverse=BFO.BFO_0000054,
        occurs_in=[],
        occurs_in_oprop=BFO.BFO_0000066,
        occurs_in_oprop_inverse=BFO.BFO_0000183,
        concretizes=[],
        concretizes_oprop=BFO.BFO_0000058,
        concretizes_oprop_inverse=BFO.BFO_0000059,
        temporal_region=None,
        temporal_region_oprop=BFO.BFO_0000199,
        spatiotemporal_region=None,
        spatiotemporal_region_oprop=BFO.BFO_0000200,
        skip_if_exists=True,
        **data_properties,
    ):
        # Init
        uri = self.add_individual_to_prefix(
            prefix, uid, label, is_a, lang, skip_if_exists, **data_properties
        )

        # Add participants
        for participant in participants:
            self.add((uri, participants_oprop, participant))
            self.add((participant, participants_oprop_inverse, uri))

        # Add realizable entities
        for realize in realizes:
            self.add((uri, realizes_oprop, realize))
            self.add((realize, realizes_oprop_inverse, uri))

        # Add occurs_in
        for oi in occurs_in:
            self.add((uri, occurs_in_oprop, oi))
            self.add((oi, occurs_in_oprop_inverse, uri))

        # Add concretizes
        for concretize in concretizes:
            self.add((uri, concretizes_oprop, concretize))
            self.add((concretize, concretizes_oprop_inverse, uri))

        # Add occupies_temporal_region
        if temporal_region:
            self.add((uri, temporal_region_oprop, temporal_region))

        # Add occupies_spatiotemporal_region
        if spatiotemporal_region:
            self.add((uri, spatiotemporal_region_oprop, spatiotemporal_region))
        return uri



import json
import re

from naas_abi_core import logger


def extract_json_from_completion(completion_text: str) -> list | dict:
    """Extract JSON object from completion text that contains markdown formatting.

    Args:
        completion_text (str): Raw completion text containing JSON in markdown format

    Returns:
        dict: Parsed JSON object
    """
    # Find JSON content between ```json and ``` markers
    json_start = completion_text.find("```json\n")
    json_end = completion_text.rfind("```")

    if json_start == -1 or json_end == -1:
        # If no markdown markers found, try parsing the whole text
        json_str = completion_text
    else:
        # Extract content between markers and clean it
        json_start += len("```json\n")
        json_str = completion_text[json_start:json_end].strip()

    # Try multiple cleanup approaches
    attempts = [
        lambda x: x,  # Original string
        lambda x: x.replace("```", "").strip(),  # Remove markdown artifacts
        lambda x: x.replace("\n", "").replace(" ", ""),  # Remove all whitespace
        lambda x: x.replace("'", '"'),  # Replace single quotes with double quotes
        lambda x: re.sub(
            r"([{,])\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*:", r'\1"\2":', x
        ),  # Quote unquoted keys
    ]

    for attempt in attempts:
        try:
            cleaned_str = attempt(json_str)
            return json.loads(cleaned_str)
        except json.JSONDecodeError as e:
            logger.debug(f"JSON parse attempt failed: {str(e)}")
            continue

    # If all attempts fail, return empty dict
    logger.error("Failed to parse JSON after all cleanup attempts")
    return {}



"""
ABI MCP Server - Lightweight HTTP-based Implementation
Exposes ABI agents as MCP tools with fast startup (no heavy imports)
"""

import asyncio
import os
import re
from typing import Any, Dict, List

import httpx
from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP

# Load environment variables from .env file if present
load_dotenv()

# Initialize FastMCP server
mcp = FastMCP("abi")

# Constants - Default to localhost for development, can be overridden by env var
ABI_API_BASE = os.environ.get("ABI_API_BASE", "http://localhost:9879")


def get_api_key() -> str:
    """Get the API key from environment variables"""
    api_key = os.environ.get("ABI_API_KEY")
    if not api_key:
        print("❌ ABI_API_KEY not found in environment")
        print("📝 Please add it to your .env file:")
        print("   ABI_API_KEY=your_key_here")
        exit(1)
    return api_key


async def fetch_openapi_spec() -> Dict[str, Any]:
    """Fetch the OpenAPI specification to discover available agents"""
    try:
        async with httpx.AsyncClient(timeout=10.0) as client:
            response = await client.get(f"{ABI_API_BASE}/openapi.json")
            response.raise_for_status()
            return response.json()
    except Exception as e:
        print(f"❌ Failed to fetch OpenAPI spec: {e}")
        return {}


def extract_agents_from_openapi(openapi_spec: Dict[str, Any]) -> List[Dict[str, str]]:
    """Extract agent information from OpenAPI specification"""
    agents = []
    paths = openapi_spec.get("paths", {})

    for path, methods in paths.items():
        # Look for agent completion endpoints
        if "/agents/" in path and path.endswith("/completion"):
            # Extract agent name from path like /agents/{agent_name}/completion
            agent_name = path.split("/agents/")[1].split("/completion")[0]

            # Get description from POST method
            post_method = methods.get("post", {})
            description = post_method.get("summary", f"{agent_name} agent completion")

            agents.append(
                {
                    "name": agent_name,
                    "description": description,
                    "function_name": agent_name_to_function_name(agent_name),
                }
            )

    return agents


def agent_name_to_function_name(agent_name: str) -> str:
    """Convert agent name to valid Python function name"""
    # Replace spaces and special chars with underscores, convert to lowercase
    function_name = re.sub(r"[^a-zA-Z0-9_]", "_", agent_name.lower())
    # Remove multiple consecutive underscores
    function_name = re.sub(r"_+", "_", function_name)
    # Remove leading/trailing underscores
    function_name = function_name.strip("_")
    # Ensure it doesn't start with a number
    if function_name and function_name[0].isdigit():
        function_name = f"agent_{function_name}"
    return function_name or "unknown_agent"


async def call_abi_agent_http(agent_name: str, prompt: str, thread_id: int = 1) -> str:
    """Call ABI agents via HTTP to avoid heavy module imports"""
    try:
        headers = {
            "Authorization": f"Bearer {get_api_key()}",
            "Content-Type": "application/json",
            "User-Agent": "ABI-MCP/1.0",
        }

        data = {"prompt": prompt, "thread_id": thread_id}

        url = f"{ABI_API_BASE}/agents/{agent_name}/completion"

        async with httpx.AsyncClient(timeout=30.0) as client:
            response = await client.post(url, json=data, headers=headers)
            response.raise_for_status()
            return response.text.strip('"')  # Remove JSON quotes

    except httpx.ConnectError:
        return f"❌ ABI API server not running at {ABI_API_BASE}. Please start it first with: uv run api"
    except httpx.TimeoutException:
        return f"⏱️ Timeout calling {agent_name} agent. The agent might be processing a complex request."
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 401:
            return (
                "🔒 Authentication failed. Check your ABI_API_KEY environment variable."
            )
        elif e.response.status_code == 404:
            return f"❓ Agent '{agent_name}' not found. Please check available agents via OpenAPI spec."
        else:
            return f"❌ HTTP {e.response.status_code} error calling {agent_name} agent: {e.response.text}"
    except Exception as e:
        return f"❌ Error calling {agent_name} agent: {str(e)}"


def create_agent_function(agent_name: str, description: str):
    """Create a dynamic agent function"""

    async def agent_function(prompt: str, thread_id: int = 1) -> str:
        return await call_abi_agent_http(agent_name, prompt, thread_id)

    # Set function metadata
    agent_function.__name__ = agent_name_to_function_name(agent_name)
    agent_function.__doc__ = f"""{description}
    
    Args:
        prompt: Your question or request for the {agent_name} agent
        thread_id: Thread ID for conversation context (default: 1)
    """

    return agent_function


async def wait_for_api():
    """Wait for the API to be available before starting"""
    max_retries = 30  # Wait up to 5 minutes (30 * 10 seconds)
    retry_delay = 10  # seconds

    for attempt in range(max_retries):
        try:
            async with httpx.AsyncClient(timeout=5.0) as client:
                response = await client.get(f"{ABI_API_BASE}")
                if response.status_code == 200:
                    print("✅ API is ready!")
                    return True
        except Exception:
            pass

        if attempt < max_retries - 1:
            print(
                f"⏳ Waiting for API to be ready... (attempt {attempt + 1}/{max_retries})"
            )
            await asyncio.sleep(retry_delay)

    print("❌ API did not become ready in time")
    return False


async def register_agents_dynamically():
    """Discover and register agents from OpenAPI specification"""
    print("🔍 Discovering agents from OpenAPI specification...")

    # Wait for API to be ready if not running locally
    if not ABI_API_BASE.startswith("http://localhost"):
        api_ready = await wait_for_api()
        if not api_ready:
            print("⚠️ API not ready, using fallback configuration")
            return

    openapi_spec = await fetch_openapi_spec()
    if not openapi_spec:
        print("⚠️ Failed to fetch OpenAPI spec, falling back to basic agents")
        return

    agents = extract_agents_from_openapi(openapi_spec)

    if not agents:
        print("⚠️ No agents found in OpenAPI spec")
        return

    print(f"📡 Found {len(agents)} agents:")

    for agent_info in agents:
        agent_name = agent_info["name"]
        description = agent_info["description"]
        function_name = agent_info["function_name"]

        print(f"  • {agent_name} -> {function_name}()")

        # Create and register the agent function
        agent_function = create_agent_function(agent_name, description)
        mcp.tool()(agent_function)

    print("✅ All agents registered successfully!")


async def setup():
    """Setup function to initialize the server"""
    # Quick startup - no heavy imports!
    print("🚀 Starting lightweight ABI MCP Server...")
    print(f"📡 Will connect to ABI API at: {ABI_API_BASE}")

    # Validate API key exists
    get_api_key()
    print("🔑 API key loaded successfully")

    # Dynamically discover and register agents
    await register_agents_dynamically()


def run():
    """Entry point for the script"""
    # Run setup first
    asyncio.run(setup())

    # Determine transport type from environment
    transport = os.environ.get("MCP_TRANSPORT", "stdio")

    if transport == "sse":
        # SSE transport for web deployment
        print(
            "🌐 Starting MCP server with SSE (Server-Sent Events) transport on port 8000"
        )
        mcp.run(transport="sse")
    elif transport == "http":
        # HTTP transport using streamable-http
        print("🌐 Starting MCP server with streamable HTTP transport")
        mcp.run(transport="streamable-http")
    else:
        # STDIO transport for local Claude Desktop integration
        print("📋 Starting MCP server with STDIO transport")
        mcp.run(transport="stdio")


if __name__ == "__main__":
    run()



import os

from naas_abi_core.services.cache.adapters.secondary.CacheFSAdapter import (
    CacheFSAdapter,
)
from naas_abi_core.services.cache.CacheService import CacheService
from naas_abi_core.utils.Storage import NoStorageFolderFound, find_storage_folder


class CacheFactory:
    @staticmethod
    def CacheFS_find_storage(
        subpath: str = "cache", needle: str = "storage"
    ) -> CacheService:
        if not subpath.startswith("cache"):
            subpath = os.path.join("cache", subpath)

        try:
            return CacheService(
                CacheFSAdapter(
                    os.path.join(find_storage_folder(os.getcwd(), needle), subpath)
                )
            )
        except NoStorageFolderFound as _:
            # Create a "storage" folder for the cache
            os.makedirs(os.path.join(os.getcwd(), "storage"), exist_ok=True)
            return CacheService(
                CacheFSAdapter(
                    os.path.join(find_storage_folder(os.getcwd(), needle), subpath)
                )
            )



from typing import TYPE_CHECKING, Literal, Union

from naas_abi_core.engine.engine_configuration.EngineConfiguration_GenericLoader import (
    GenericLoader,
)
from naas_abi_core.engine.engine_configuration.EngineConfiguration_ObjectStorageService import (
    ObjectStorageServiceConfiguration,
)
from naas_abi_core.engine.engine_configuration.utils.PydanticModelValidator import (
    pydantic_model_validator,
)
from naas_abi_core.services.triple_store.TripleStorePorts import ITripleStorePort
from naas_abi_core.services.triple_store.TripleStoreService import TripleStoreService
from pydantic import BaseModel, model_validator
from typing_extensions import Self

# Only import for type checking, not at runtime
if TYPE_CHECKING:
    pass


class OxigraphAdapterConfiguration(BaseModel):
    """Oxigraph adapter configuration.

    triple_store_adapter:
      adapter: "oxigraph"
      config:
        oxigraph_url: "http://localhost:7878"
        timeout: 60
    """
    oxigraph_url: str = "http://localhost:7878"
    timeout: int = 60


class AWSNeptuneAdapterConfiguration(BaseModel):
    """AWS Neptune adapter configuration.

    triple_store_adapter:
      adapter: "aws_neptune"
      config:
        aws_region_name: "eu-west-3"
        aws_access_key_id: "{{ secret.AWS_ACCESS_KEY_ID }}"
        aws_secret_access_key: "{{ secret.AWS_SECRET_ACCESS_KEY }}"
        db_instance_identifier: "{{ secret.DB_INSTANCE_IDENTIFIER }}"
    """
    aws_region_name: str
    aws_access_key_id: str
    aws_secret_access_key: str
    db_instance_identifier: str


class AWSNeptuneSSHTunnelAdapterConfiguration(AWSNeptuneAdapterConfiguration):
    """AWS Neptune SSH tunnel adapter configuration.

    triple_store_adapter:
      adapter: "aws_neptune_sshtunnel"
      config:
        aws_region_name: "eu-west-3"
        aws_access_key_id: "{{ secret.AWS_ACCESS_KEY_ID }}"
        aws_secret_access_key: "{{ secret.AWS_SECRET_ACCESS_KEY }}"
        db_instance_identifier: "{{ secret.DB_INSTANCE_IDENTIFIER }}"   
        bastion_host: "bastion.example.com"
        bastion_port: 22
        bastion_user: "ubuntu"
        bastion_private_key: "{{ secret.BASTION_PRIVATE_KEY }}"
    """
    bastion_host: str
    bastion_port: int
    bastion_user: str
    bastion_private_key: str


class TripleStoreAdapterFilesystemConfiguration(BaseModel):
    """Filesystem adapter configuration.

    triple_store_adapter:
      adapter: "fs"
      config:
        store_path: "storage/triplestore"
        triples_path: "triples"
    """
    store_path: str
    triples_path: str = "triples"


class TripleStoreAdapterObjectStorageConfiguration(BaseModel):
    """Object storage adapter configuration.

    triple_store_adapter:
      adapter: "object_storage"
      config:
        object_storage_service: *object_storage_service
        triples_prefix: "triples"
    """
    object_storage_service: ObjectStorageServiceConfiguration
    triples_prefix: str = "triples"


class TripleStoreAdapterConfiguration(GenericLoader):
    adapter: Literal[
        "oxigraph",
        "aws_neptune_sshtunnel",
        "aws_neptune",
        "fs",
        "object_storage",
        "custom",
    ]
    config: (
        Union[
            OxigraphAdapterConfiguration,
            AWSNeptuneAdapterConfiguration,
            AWSNeptuneSSHTunnelAdapterConfiguration,
            TripleStoreAdapterFilesystemConfiguration,
            TripleStoreAdapterObjectStorageConfiguration,
            dict,
        ]
        | None
    ) = None

    @model_validator(mode="after")
    def validate_adapter(self) -> Self:
        if self.adapter != "custom":
            assert self.config is not None, (
                "config is required if adapter is not custom"
            )

        if self.adapter == "fs":
            pydantic_model_validator(
                TripleStoreAdapterFilesystemConfiguration,
                self.config,
                "Invalid configuration for services.triple_store.triple_store_adapter 'fs' adapter",
            )
        if self.adapter == "object_storage":
            pydantic_model_validator(
                TripleStoreAdapterObjectStorageConfiguration,
                self.config,
                "Invalid configuration for services.triple_store.triple_store_adapter 'object_storage' adapter",
            )
        if self.adapter == "oxigraph":
            pydantic_model_validator(
                OxigraphAdapterConfiguration,
                self.config,
                "Invalid configuration for services.triple_store.triple_store_adapter 'oxigraph' adapter",
            )
        if self.adapter == "aws_neptune":
            pydantic_model_validator(
                AWSNeptuneAdapterConfiguration,
                self.config,
                "Invalid configuration for services.triple_store.triple_store_adapter 'aws_neptune' adapter",
            )
        if self.adapter == "aws_neptune_sshtunnel":
            pydantic_model_validator(
                AWSNeptuneSSHTunnelAdapterConfiguration,
                self.config,
                "Invalid configuration for services.triple_store.triple_store_adapter 'aws_neptune_sshtunnel' adapter",
            )

        return self

    def load(self) -> ITripleStorePort:
        if self.adapter != "custom":
            assert self.config is not None, "config is required"

            arguments = (
                self.config.model_dump()
                if not isinstance(self.config, dict)
                else self.config
            )

            # Lazy import: only import the adapter that's actually configured
            if self.adapter == "oxigraph":
                from naas_abi_core.services.triple_store.adaptors.secondary.Oxigraph import (
                    Oxigraph,
                )

                OxigraphAdapterConfiguration.model_validate(arguments)

                return Oxigraph(**arguments)
            elif self.adapter == "aws_neptune":
                from naas_abi_core.services.triple_store.adaptors.secondary.AWSNeptune import (
                    AWSNeptune,
                )

                AWSNeptuneAdapterConfiguration.model_validate(arguments)

                return AWSNeptune(**arguments)
            elif self.adapter == "aws_neptune_sshtunnel":
                from naas_abi_core.services.triple_store.adaptors.secondary.AWSNeptune import (
                    AWSNeptuneSSHTunnel,
                )

                AWSNeptuneSSHTunnelAdapterConfiguration.model_validate(arguments)

                return AWSNeptuneSSHTunnel(**arguments)
            elif self.adapter == "fs":
                from naas_abi_core.services.triple_store.adaptors.secondary.TripleStoreService__SecondaryAdaptor__Filesystem import (
                    TripleStoreService__SecondaryAdaptor__Filesystem,
                )

                TripleStoreAdapterFilesystemConfiguration.model_validate(arguments)

                return TripleStoreService__SecondaryAdaptor__Filesystem(**arguments)
            elif self.adapter == "object_storage":
                from naas_abi_core.services.triple_store.adaptors.secondary.TripleStoreService__SecondaryAdaptor__ObjectStorage import (
                    TripleStoreService__SecondaryAdaptor__ObjectStorage,
                )

                return TripleStoreService__SecondaryAdaptor__ObjectStorage(**arguments)
            else:
                raise ValueError(f"Adapter {self.adapter} not supported")
        else:
            return super().load()


class TripleStoreServiceConfiguration(BaseModel):
    triple_store_adapter: TripleStoreAdapterConfiguration

    def load(self) -> TripleStoreService:
        return TripleStoreService(triple_store_adapter=self.triple_store_adapter.load())



import pytest


@pytest.fixture
def model():
    import os

    from dotenv import load_dotenv
    from langchain_openai import ChatOpenAI

    load_dotenv()

    return ChatOpenAI(
        model="gpt-4o", temperature=0, api_key=os.environ.get("OPENAI_API_KEY")
    )


def test_no_tools_no_agents(model):
    from langchain_core.messages import AIMessage
    from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration

    agent = Agent(
        name="Test Agent",
        description="A test agent",
        chat_model=model,
        tools=[],
        agents=[],
        configuration=AgentConfiguration(
            system_prompt="You are an LLM running in a test environment. You must output '42' and nothing else."
        ),
    )

    for _, chunk in agent.stream("Hello, world!"):
        assert "call_model" in chunk
        assert "messages" in chunk["call_model"]
        assert len(chunk["call_model"]["messages"]) == 1
        assert isinstance(chunk["call_model"]["messages"][0], AIMessage)
        assert chunk["call_model"]["messages"][0].content == "42"


def test_tools_no_agents(model):
    from langchain_core.tools import tool
    from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration

    @tool
    def test_tool(input: str) -> str:
        """test_tool

        Args:
            input (str): The input to the tool

        Returns:
            str: Return the input
        """
        assert input == "Hello, world!"
        return input

    agent = Agent(
        name="Test Agent",
        description="A test agent",
        chat_model=model,
        tools=[test_tool],
        agents=[],
        configuration=AgentConfiguration(
            system_prompt="You are an LLM running in a test environment. You must call the tool `test_tool` with the argument `Hello, world!` and format the output as `Tool Response: <result>`. It is very important that you reformat the output and not return only the tool response."
        ),
    )

    chunks = []

    for _, chunk in agent.stream("Hello, world!"):
        chunks.append(chunk)


# This test is not consistent.
# def test_agents_no_tools(model):
#     from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration, AgentSharedState
#     from queue import Queue
#     import json
#     import pydash as pd

#     queue = Queue()

#     agent = Agent(
#         name="Test Agent",
#         description="A test agent",
#         chat_model=model,
#         tools=[
#             Agent(
#                 name="Greeter",
#                 description="A greeter agent",
#                 chat_model=model,
#                 tools=[],
#                 agents=[],
#                 configuration=AgentConfiguration(
#                     system_prompt="You are a greeter agent. You must greet the user with the name they provide in the form 'Hello, <name>!' and nothing else."
#                 ),
#                 event_queue=queue,
#             )
#         ],
#         agents=[],
#         configuration=AgentConfiguration(
#             system_prompt="The user will send you his name. You must call the greeter agent and output it's result and nothing else. You must call it only once."
#         ),
#         event_queue=queue,
#     )

#     chunks = []

#     for _, chunk in agent.stream("ABI"):
#         chunks.append(chunk)
#         if 'call_model' in chunk:
#             print(f'\n\t Call Model: {type(chunk)}')
#             print(pd.get(chunk, "call_model.messages[-1].content"))
#             print('\n')
#         elif 'call_tools' in chunk:
#             print(f'\n\t Call Tools: {type(chunk)}')
#             print(pd.get(chunk, "call_tools.messages[-1].content"))
#             print('\n')
#         else:
#             print(f'\n\t Unknown: {type(chunk)}')
#             print(chunk)
#             print('\n')
#         assert len(chunks) <= 5

#     assert len(chunks) == 5
#     assert list(chunks[-1].values())[0]['messages'][-1].content == "Hello, ABI!"


def test_agent_duplication(model):
    from queue import Queue

    from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration

    queue = Queue()

    sub_agent = Agent(
        name="Sub agent",
        description="Sub agent",
        chat_model=model,
        tools=[],
        configuration=AgentConfiguration(system_prompt=""),
        event_queue=queue,
    )

    first_agent = Agent(
        name="Test Agent",
        description="A test agent",
        chat_model=model,
        tools=[sub_agent],
        agents=[],
        configuration=AgentConfiguration(system_prompt=""),
        event_queue=queue,
    )

    duplicated_agent = first_agent.duplicate()

    assert id(duplicated_agent) != id(first_agent)
    assert id(duplicated_agent.agents[0]) != id(first_agent.agents[0])


# def test_agent_stream_invoke(model):
#     from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration

#     agent = Agent(
#         name="Greeting Agent",
#         description="A Greeting agent",
#         chat_model=model,
#         tools=[],
#         agents=[],
#         configuration=AgentConfiguration(system_prompt="You are a Greeting agent"),
#     )

#     events = []
#     for event in agent.stream_invoke("My name is ABI"):
#         events.append(event)

#     assert len(events) == 2
#     assert events[0]["event"] == "message"
#     assert events[1]["event"] == "done"


def test_agent_one_tool_agent_response(model):
    from langchain_core.tools import tool
    from naas_abi_core.services.agent.Agent import Agent, AgentConfiguration

    @tool
    def add_one(input: int) -> int:
        """Add one to the input"""
        return input + 1

    agent = Agent(
        name="Test Agent",
        description="A test agent",
        chat_model=model,
        tools=[add_one],
        configuration=AgentConfiguration(
            system_prompt="""You are a test agent. You must add one to the input and return the result.
        Your output must be like:
        
        
        <user_input> + one = result
        """
        ),
    )

    chunks = []
    for _, chunk in agent.stream("42"):
        chunks.append(chunk)
        print(chunk)

    assert len(chunks) == 3
    assert "call_model" in chunks[-1]
    assert chunks[-1]["call_model"]["messages"][0].content == "42 + one = 43"



import json
from datetime import datetime
from io import BytesIO
from typing import Dict, Tuple

import pandas as pd
import yaml
from rdflib import Graph

from naas_abi_core import logger
from naas_abi_core.services.object_storage.ObjectStoragePort import IObjectStorageDomain


class StorageUtils:
    __storage_service: IObjectStorageDomain

    def __init__(self, storage_service: IObjectStorageDomain):
        self.__storage_service = storage_service

    def __make_copy(
        self, dir_path: str, file_name: str, content: bytes
    ) -> Tuple[str, str]:
        """
        Make a copy of a file in storage with timestamp in the name.
        """
        try:
            file_name = f"{datetime.now().strftime('%Y%m%dT%H%M%S')}_{file_name}"
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=content
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error making copy in {dir_path}: {e}")
            return dir_path, file_name

    def get_text(
        self, dir_path: str, file_name: str, encoding: str = "utf-8"
    ) -> str | None:
        """
        Get a text file from storage.
        """
        try:
            content = self.__storage_service.get_object(dir_path, file_name)
            return content.decode(encoding)
        except Exception as e:
            logger.warning(f"Error getting text from {dir_path}: {e}")
            return None

    def save_text(
        self,
        text: str,
        dir_path: str,
        file_name: str,
        encoding: str = "utf-8",
        copy: bool = True,
    ) -> Tuple[str, str]:
        """
        Save a text file to storage.
        """
        try:
            content = text.encode(encoding)
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=content
            )
            if copy:
                self.__make_copy(dir_path, file_name, content)
            logger.debug(
                f"[save_text] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving text to {dir_path}: {e}")
            return dir_path, file_name

    def get_image(self, dir_path: str, file_name: str) -> bytes | None:
        """
        Get an image from storage.
        """
        try:
            return self.__storage_service.get_object(dir_path, file_name)
        except Exception as e:
            logger.warning(f"Error getting image from {dir_path}: {e}")
            return None

    def save_image(
        self, image: bytes, dir_path: str, file_name: str, copy: bool = True
    ) -> Tuple[str, str]:
        """
        Save an image to storage.
        """
        try:
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=image
            )
            if copy:
                self.__make_copy(dir_path, file_name, image)
            logger.debug(
                f"[save_image] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving image to {dir_path}: {e}")
            return dir_path, file_name

    def get_csv(
        self,
        dir_path: str,
        file_name: str,
        sep: str = ";",
        decimal: str = ",",
        encoding: str = "utf-8",
    ) -> pd.DataFrame:
        """
        Get a CSV file from storage.
        """
        try:
            file_content = self.__storage_service.get_object(dir_path, file_name)
            # Create a BytesIO object to avoid file name length issues
            from io import BytesIO

            csv_buffer = BytesIO(file_content)
            return pd.read_csv(csv_buffer, sep=sep, decimal=decimal, encoding=encoding)
        except Exception as e:
            logger.warning(f"Error getting CSV file from {dir_path}: {e}")
            return pd.DataFrame()

    def save_csv(
        self,
        data: pd.DataFrame,
        dir_path: str,
        file_name: str,
        sep: str = ";",
        decimal: str = ",",
        encoding: str = "utf-8",
        copy: bool = True,
    ) -> Tuple[str, str]:
        """
        Save a CSV file to storage.
        """
        try:
            self.__storage_service.put_object(
                prefix=dir_path,
                key=file_name,
                content=data.to_csv(
                    index=False, encoding=encoding, sep=sep, decimal=decimal
                ).encode(encoding),
            )
            if copy:
                self.__make_copy(
                    dir_path,
                    file_name,
                    data.to_csv(
                        index=False, encoding=encoding, sep=sep, decimal=decimal
                    ).encode(encoding),
                )
            logger.debug(
                f"[save_csv] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving CSV file to {dir_path}: {e}")
            return dir_path, file_name

    def get_excel(
        self,
        dir_path: str,
        file_name: str,
        sheet_name: str,
        skiprows: int = 0,
        usecols: list | None = None,
    ) -> pd.DataFrame:
        """
        Get an Excel file from storage.
        """
        try:
            file_content = BytesIO(
                self.__storage_service.get_object(dir_path, file_name)
            )
            return pd.read_excel(
                file_content, sheet_name=sheet_name, skiprows=skiprows, usecols=usecols
            )
        except Exception as e:
            logger.warning(f"Error getting Excel file from {dir_path}: {e}")
            return pd.DataFrame()

    def save_excel(
        self,
        data: pd.DataFrame,
        dir_path: str,
        file_name: str,
        sheet_name: str,
        copy: bool = True,
    ) -> Tuple[str, str]:
        """
        Save an Excel file to storage.
        """
        try:
            excel_buffer = BytesIO()
            data.to_excel(excel_buffer, index=False, sheet_name=sheet_name)
            excel_buffer.seek(0)
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=excel_buffer.getvalue()
            )
            if copy:
                self.__make_copy(dir_path, file_name, excel_buffer.getvalue())
            logger.debug(
                f"[save_excel] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving Excel file to {dir_path}: {e}")
            return dir_path, file_name

    def get_json(self, dir_path: str, file_name: str) -> Dict:
        """
        Get JSON data from storage.
        """
        try:
            file_content = self.__storage_service.get_object(
                dir_path, file_name
            ).decode("utf-8")
            data = json.loads(file_content)
            return data
        except Exception as e:
            logger.warning(f"Error getting JSON data from {dir_path}: {e}")
            return {}

    def save_json(
        self, data: dict | list, dir_path: str, file_name: str, copy: bool = True
    ) -> Tuple[str, str]:
        """
        Save JSON data to storage.
        """
        try:
            self.__storage_service.put_object(
                prefix=dir_path,
                key=file_name,
                content=json.dumps(data, indent=4, ensure_ascii=False).encode("utf-8"),
            )
            if copy:
                # Only create a timestamped copy in storage, not local filesystem
                self.__make_copy(
                    dir_path,
                    file_name,
                    json.dumps(data, indent=4, ensure_ascii=False).encode("utf-8"),
                )
            logger.debug(
                f"[save_json] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving JSON data to {dir_path}: {e}")
            return dir_path, file_name

    def get_yaml(self, dir_path: str, file_name: str) -> Dict:
        """
        Get YAML data from storage.
        """
        try:
            file_content = self.__storage_service.get_object(
                dir_path, file_name
            ).decode("utf-8")
            data = yaml.safe_load(file_content)
            return data if data is not None else {}
        except Exception as e:
            logger.debug(f"Error getting YAML data from {dir_path}: {e}")
            return {}

    def save_yaml(
        self, data: dict | list, dir_path: str, file_name: str, copy: bool = True
    ) -> Tuple[str, str]:
        """
        Save YAML data to storage.
        """
        try:
            self.__storage_service.put_object(
                prefix=dir_path,
                key=file_name,
                content=yaml.dump(
                    data, default_flow_style=False, allow_unicode=True, sort_keys=False
                ).encode("utf-8"),
            )
            if copy:
                self.__make_copy(
                    dir_path,
                    file_name,
                    yaml.dump(
                        data,
                        default_flow_style=False,
                        allow_unicode=True,
                        sort_keys=False,
                    ).encode("utf-8"),
                )
            logger.debug(
                f"[save_yaml] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving YAML data to {dir_path}: {e}")
            return dir_path, file_name

    def get_triples(
        self, dir_path: str, file_name: str, format: str = "turtle"
    ) -> Graph:
        """
        Get Turtle data from storage and return as RDFlib Graph.
        """
        try:
            file_content = self.__storage_service.get_object(
                dir_path, file_name
            ).decode("utf-8")
            graph = Graph()
            graph.parse(data=file_content, format=format)
            return graph
        except Exception as e:
            logger.warning(f"Error getting triples from {dir_path}: {e}")
            return Graph()

    def save_triples(
        self,
        graph: Graph,
        dir_path: str,
        file_name: str,
        format: str = "turtle",
        copy: bool = True,
    ) -> Tuple[str, str]:
        """
        Save RDFlib Graph data to storage in Turtle format.
        """
        try:
            turtle_content = graph.serialize(format=format, sort=False)
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=turtle_content.encode("utf-8")
            )
            if copy:
                self.__make_copy(dir_path, file_name, turtle_content.encode("utf-8"))
            logger.debug(
                f"[save_triples] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving triples to {dir_path}: {e}")
            return dir_path, file_name

    def get_powerpoint_presentation(self, dir_path: str, file_name: str) -> BytesIO:
        """
        Get a PowerPoint presentation from storage.

        Args:
            output_dir: The directory where the presentation is stored.
            file_name: The name of the presentation file.

        Returns:
            BytesIO: A BytesIO object containing the presentation data.
        """
        try:
            data = self.__storage_service.get_object(dir_path, file_name)
            byte_stream = BytesIO(data)
            byte_stream.seek(0)
            return byte_stream
        except Exception as e:
            logger.warning(
                f"Error getting PowerPoint presentation from {dir_path}: {e}"
            )
            return BytesIO()

    def save_powerpoint_presentation(
        self, presentation, dir_path: str, file_name: str, copy: bool = True
    ) -> Tuple[str, str]:
        """
        Save a PowerPoint presentation to a file and create an asset in Naas.

        Args:
            presentation: The presentation to save.
            output_dir: The directory to save the presentation to.
            file_name: The name of the file to save the presentation to.
            naas_integration: The Naas integration to use to create the asset.

        Returns:
            str: The URL of the asset.
        """
        try:
            # Create byte stream
            byte_stream = BytesIO()
            presentation.save(byte_stream)
            byte_stream.seek(0)
            self.__storage_service.put_object(
                prefix=dir_path, key=file_name, content=byte_stream.getvalue()
            )
            if copy:
                self.__make_copy(dir_path, file_name, byte_stream.getvalue())
            logger.debug(
                f"[save_powerpoint_presentation] File successfully written to storage: {dir_path}/{file_name}"
            )
            return dir_path, file_name
        except Exception as e:
            logger.error(f"Error saving PowerPoint presentation to {dir_path}: {e}")
            return dir_path, file_name
